#include<mpi.h>
#include"main.h"
int main(int argc, char** argv)
{
	int ProcessRank;
	int NumberOfProcesses;

	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &ProcessRank);
	MPI_Comm_size(MPI_COMM_WORLD, &NumberOfProcesses);

	// NOTE: Check if the workload is divisible by number of workers
	if(SIZE%NumberOfProcesses != 0)
	{
		printf("Workload is not divisible by %d Processes", NumberOfProcesses);
		MPI_Finalize();
		return(0);
	}

	int BlockCount = 2;
	int BlockLengths[2] = {1, 1};
	MPI_Aint Offsets[2] = {(MPI_Aint)OFFSET_OF(v2, x),
			       (MPI_Aint)OFFSET_OF(v2, y)};
	MPI_Datatype Types[2] = {MPI_FLOAT, MPI_FLOAT};
	MPI_Datatype MPI_V2_TEMP;
	    
	MPI_Type_struct(BlockCount, BlockLengths, Offsets, Types, &MPI_V2_TEMP);

	// Account for Padding
	MPI_Datatype MPI_V2;
	MPI_Type_create_resized(MPI_V2_TEMP,
				Offsets[0],
				(MPI_Aint)sizeof(v2),
				&MPI_V2);
	MPI_Type_free(&MPI_V2_TEMP);
	
	MPI_Type_commit ( &MPI_V2);

	int CellBlockCount = 3;
	int CellBlockLengths[3] = {1, 1, 1};
	MPI_Aint CellOffsets[3] = {(MPI_Aint)OFFSET_OF(sim_cell, ParticleCount),
				   (MPI_Aint)OFFSET_OF(sim_cell, Velocity),
	                           (MPI_Aint)OFFSET_OF(sim_cell, Type)};
	MPI_Datatype CellTypes[3] = {MPI_FLOAT, MPI_V2, MPI_INT};
	MPI_Datatype MPI_SIM_CELL_TEMP;
	    
	MPI_Type_struct(CellBlockCount, CellBlockLengths, CellOffsets, CellTypes, &MPI_SIM_CELL_TEMP);

	// Account for Padding
	MPI_Datatype MPI_SIM_CELL;
	MPI_Type_create_resized(MPI_SIM_CELL_TEMP,
				Offsets[0],
				(MPI_Aint)sizeof(sim_cell),
				&MPI_SIM_CELL);
	MPI_Type_free(&MPI_SIM_CELL_TEMP);
	
	MPI_Type_commit ( &MPI_SIM_CELL);

	
	int RowsPerProcess = SIZE/NumberOfProcesses;

	sim_cell *Array;
	sim_cell *TempArray;
	sim_cell *WorkArray;
	if(ProcessRank == 0)
	  {
	    Array = malloc(sizeof(sim_cell)*(SIZE+2)*SIZE);
	    // NOTE: Some Testvalues
	    for (int y = 0; y < SIZE+2 ; y++)
	      {
	        for (int x = 0; x < SIZE; x++)
		  {
	            sim_cell *Cell = (Array + (y*SIZE)+x);
	            Cell->ParticleCount = 1000;
	            Cell->Velocity = V2(0,0);
	            Cell->Type = 0;
		  }
	      }
	    /*	    
	    sim_cell *Cell_5_5 = (Array+(5*SIZE)+5);
	    sim_cell *Cell_5_4 = (Array+(4*SIZE)+5);
	    sim_cell *Cell_6_4 = (Array+(4*SIZE)+6);
	    Cell_5_5->ParticleCount = 10000.0f;
	    Cell_5_4->ParticleCount = 15000.0f;
	    Cell_6_4->ParticleCount = 500.0f;
	    (Array+(10*SIZE)+13)->Type = SimCellType_Obsticle;
	    (Array+(11*SIZE)+12)->Type = SimCellType_Obsticle;
	    (Array+(12*SIZE)+11)->Type = SimCellType_Obsticle;
	    (Array+(13*SIZE)+10)->Type = SimCellType_Obsticle;	
	    (Array+(14*SIZE)+9)->Type = SimCellType_Obsticle;
	    (Array+(15*SIZE)+8)->Type = SimCellType_Obsticle;
	    (Array+(16*SIZE)+7)->Type = SimCellType_Obsticle;
	    (Array+(17*SIZE)+6)->Type = SimCellType_Obsticle;	
	    (Array+(18*SIZE)+5)->Type = SimCellType_Obsticle;
	    (Array+(19*SIZE)+4)->Type = SimCellType_Obsticle;
	    */

	    (Array+(28*SIZE)+22)->Type = SimCellType_Obsticle;
	    (Array+(27*SIZE)+22)->Type = SimCellType_Obsticle;
	    (Array+(26*SIZE)+22)->Type = SimCellType_Obsticle;
	    (Array+(25*SIZE)+22)->Type = SimCellType_Obsticle;	
	    (Array+(24*SIZE)+22)->Type = SimCellType_Obsticle;
	    (Array+(23*SIZE)+22)->Type = SimCellType_Obsticle;
	    (Array+(22*SIZE)+22)->Type = SimCellType_Obsticle;
	    
	    (Array+(28*SIZE)+28)->Type = SimCellType_Obsticle;
	    (Array+(27*SIZE)+28)->Type = SimCellType_Obsticle;
	    (Array+(26*SIZE)+28)->Type = SimCellType_Obsticle;
	    (Array+(25*SIZE)+28)->Type = SimCellType_Obsticle;	
	    (Array+(24*SIZE)+28)->Type = SimCellType_Obsticle;
	    (Array+(23*SIZE)+28)->Type = SimCellType_Obsticle;
	    (Array+(22*SIZE)+28)->Type = SimCellType_Obsticle;

	    (Array+(22*SIZE)+28)->Type = SimCellType_Obsticle;
	    (Array+(22*SIZE)+27)->Type = SimCellType_Obsticle;
	    (Array+(22*SIZE)+26)->Type = SimCellType_Obsticle;
	    (Array+(22*SIZE)+25)->Type = SimCellType_Obsticle;	
	    (Array+(22*SIZE)+24)->Type = SimCellType_Obsticle;
	    (Array+(22*SIZE)+23)->Type = SimCellType_Obsticle;
	    (Array+(22*SIZE)+22)->Type = SimCellType_Obsticle;
	    
	    (Array+(28*SIZE)+28)->Type = SimCellType_Obsticle;
	    (Array+(28*SIZE)+27)->Type = SimCellType_Obsticle;
	    //(Array+(28*SIZE)+26)->Type = SimCellType_Obsticle;
	    //(Array+(28*SIZE)+25)->Type = SimCellType_Obsticle;	
	    (Array+(28*SIZE)+24)->Type = SimCellType_Obsticle;
	    (Array+(28*SIZE)+23)->Type = SimCellType_Obsticle;
	    (Array+(28*SIZE)+22)->Type = SimCellType_Obsticle;

	    
	    (Array+(25*SIZE)+24)->ParticleCount = 5500;
	    (Array+(25*SIZE)+25)->ParticleCount = 5500;
	    (Array+(25*SIZE)+26)->ParticleCount = 5500;
	    (Array+(26*SIZE)+24)->ParticleCount = 5000;	
	    (Array+(26*SIZE)+25)->ParticleCount = 5500;
	    (Array+(26*SIZE)+26)->ParticleCount = 5500;
	    (Array+(24*SIZE)+24)->ParticleCount = 5500;
	    (Array+(24*SIZE)+25)->ParticleCount = 5500;
	    (Array+(24*SIZE)+26)->ParticleCount = 5500;

	    (Array+(23*SIZE)+25)->ParticleCount = 2000;
	    (Array+(27*SIZE)+25)->ParticleCount = 2000;
	    (Array+(25*SIZE)+23)->ParticleCount = 2000;
	    (Array+(25*SIZE)+27)->ParticleCount = 2000;
	}
	#if 0
	else
	{
		Array = malloc(sizeof(int)*(RowsPerProcess+2)*SIZE);
	}
#endif
	WorkArray = malloc(sizeof(sim_cell)*(RowsPerProcess+2)*SIZE);;
	TempArray = malloc(sizeof(sim_cell)*RowsPerProcess*SIZE);
	//memset(TempArray, 0, sizeof(sim_cell)*RowsPerProcess*SIZE);

	int SCounts[NumberOfProcesses];
	for(int i = 0; i < NumberOfProcesses; i++)
	{
		SCounts[i] = (RowsPerProcess+2)*SIZE;
	}

	int SDisplacements[NumberOfProcesses];
	for(int i = 0; i < NumberOfProcesses; i++)
	{
		SDisplacements[i] = i*RowsPerProcess*SIZE;
	}

	int RCounts[NumberOfProcesses];
	for(int i = 0; i < NumberOfProcesses; i++)
	{
		RCounts[i] = RowsPerProcess*SIZE;
	}

	int RDisplacements[NumberOfProcesses];
	for(int i = 0; i < NumberOfProcesses; i++)
	{
		RDisplacements[i] = ((i*RowsPerProcess)+1)*SIZE;
	}

	for(int Iteration = 0; Iteration < 100; Iteration++)
	  {
	MPI_Scatterv(Array, SCounts, SDisplacements, MPI_SIM_CELL, WorkArray, SIZE*(RowsPerProcess+2), MPI_SIM_CELL, 0, MPI_COMM_WORLD);

	int PrintVector = 0;
	
	// TODO: DO Work
 	for (int y = 1; y < RowsPerProcess+1 ; y++)
	{
		for (int x = 0; x < SIZE; x++)
		{
		        sim_cell *Cur = WorkArray + (y*SIZE) + x;
			sim_cell *TempCur = TempArray + ((y-1)*SIZE) + x;
			CopyCell(Cur, TempCur);
		        for(int xOffset = -1; xOffset <= 1 ; xOffset ++)
			{
				for(int yOffset = -1; yOffset <= 1 ; yOffset ++)
				{
					if(!(x+xOffset < 0 || x+xOffset >= SIZE) &&
					   !(y+yOffset < 0 || y+yOffset > RowsPerProcess+1) &&
					   !(y+yOffset+(ProcessRank*RowsPerProcess) < 1 || y+yOffset+(ProcessRank*RowsPerProcess) > SIZE))
					{
						sim_cell *NeighbourCell = WorkArray + (SIZE*(y+yOffset) + (x+xOffset));

											// Calculate the "pressure gradient"
						v2 Acceleration1 = {0.0f,0.0f};
						CalcPressureAcceleration(&Acceleration1, WorkArray, ProcessRank, RowsPerProcess, x+xOffset, y+yOffset, -1, 0);
						CalcPressureAcceleration(&Acceleration1, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, -1, -1);
						CalcPressureAcceleration(&Acceleration1, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, 0, -1);	
						v2 Drag1 = scale((-1.0f/3.0f), NeighbourCell->Velocity);
						Acceleration1 = add(Drag1, Acceleration1);
				
						Update(Acceleration1, x, y, xOffset, yOffset, WorkArray, TempArray, PrintVector);
				
						v2 Acceleration2 = {0.0f,0.0f};
						CalcPressureAcceleration(&Acceleration2, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, 0, -1);				
						CalcPressureAcceleration(&Acceleration2, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, 1, -1);
						CalcPressureAcceleration(&Acceleration2, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, 1, 0);
						v2 Drag2 = scale((-1.0f/3.0f), NeighbourCell->Velocity);
						Acceleration2 = add(Drag2, Acceleration2);
				
						Update(Acceleration2, x, y, xOffset, yOffset, WorkArray, TempArray, PrintVector);
				
						v2 Acceleration3 = {0.0f,0.0f};
						CalcPressureAcceleration(&Acceleration3, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, 1, 0);
						CalcPressureAcceleration(&Acceleration3, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, 1, 1);
						CalcPressureAcceleration(&Acceleration3, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, 0, 1);
						v2 Drag3 = scale((-1.0f/3.0f), NeighbourCell->Velocity);
						Acceleration3 = add(Drag3, Acceleration3);
				
						Update(Acceleration3, x, y, xOffset, yOffset, WorkArray, TempArray, PrintVector);
				
						v2 Acceleration4 = {0.0f,0.0f};
						CalcPressureAcceleration(&Acceleration4, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, 0, 1);
						CalcPressureAcceleration(&Acceleration4, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, -1, 1);
						CalcPressureAcceleration(&Acceleration4, WorkArray, ProcessRank, RowsPerProcess,x+xOffset, y+yOffset, -1, 0);
						v2 Drag4 = scale((-1.0f/3.0f), NeighbourCell->Velocity);				
						Acceleration4 = add(Drag4, Acceleration4);
				
						Update(Acceleration4, x, y, xOffset, yOffset, WorkArray, TempArray, PrintVector);

					}
				}
			}
		}
	}
	
	MPI_Gatherv(TempArray, SIZE*RowsPerProcess, MPI_SIM_CELL, Array, RCounts, RDisplacements, MPI_SIM_CELL, 0, MPI_COMM_WORLD);

	if(ProcessRank == 0)
	{
		PrintMatrix(Array, 1, SIZE+1);
	}
	  }
	free(Array);
	free(TempArray);
	free(WorkArray);
	
	MPI_Type_free(&MPI_SIM_CELL);
	MPI_Type_free(&MPI_V2);
	MPI_Finalize();
	return(0);
}

void PrintMatrix(sim_cell *Matrix, int From , int To)
{
	// NOTE: Print DATA
	for (int y = From; y < To ; y++)
	{
		for (int x = 0; x < SIZE; x++)
		{
			sim_cell *Cell = (Matrix + (y*SIZE)+x);
			printf("%4.2f ", (Cell->Type == SimCellType_Space)?Cell->ParticleCount:0);
		}
		printf("\n");
	}
	printf("\n\n");
}

void CopyCell(sim_cell *Source, sim_cell *Dest)
{
	Dest->ParticleCount = Source->ParticleCount;
	Dest->Velocity = Source->Velocity;
	Dest->Type = Source->Type;
}

void CalcPressureAcceleration(v2 *Acceleration, sim_cell *Cells, int ProcessRank, int RowsPerProcess,int x, int y, int xOffset, int yOffset)
{
	if(x+xOffset >= 0 && x+xOffset <SIZE && y+yOffset >= 0 && y+yOffset< SIZE)
	{
		sim_cell *Cell = Cells + SIZE * (y) + (x);
		sim_cell *Neighbor = Cells + SIZE * (y+yOffset) + (x+xOffset);
		// TODO: Think aboud particle bouncing of (mirror the velocity or somethink like that)
		// TODO: Also should i check for diagonals. Rigth now particles con pass through diagonal "walls"
		sim_cell *DNX = Cells + (SIZE*(y+yOffset))+x;
		sim_cell *DNY = Cells + (SIZE*y) + x + xOffset;
		// NOTE: The cell must be free and the diagonal must have at least one free one.
		if(Neighbor->Type == SimCellType_Space && (DNX->Type == SimCellType_Space || DNY->Type == SimCellType_Space))
		{
			float delta = Cell->ParticleCount - Neighbor->ParticleCount;

			if (delta > 0)
			{								
				v2 PressureAcc = scale((0.001f)*delta, V2(xOffset, yOffset));
				*Acceleration = add(*Acceleration, PressureAcc);
			}
		}
	}
}

// TODO: PrintVector should not be part of this function
void Update(v2 Acceleration, int x, int y, int xo, int yo, sim_cell *Cells, sim_cell *TempCells, bool PrintVector)
{	
        // The Current Cell (if x and y != 0 then the neighbour)
        sim_cell *Cell = Cells + ((y+yo)*SIZE) + x+xo;

	v2 NewVelocity = add(Cell->Velocity, Acceleration);
	float Percentage = MapIntoRangeClamped01(0, SIZE, Length(NewVelocity))/4.0f;

	sim_cell *NewCell = TempCells + (y-1)*SIZE + (x);

	sim_cell *Neighbor;
	sim_cell *NewNeighbor;

	v2 dPos = add(scale((1.0f/2.0f), Acceleration), NewVelocity); 
												      
	int xOffset = 0, yOffset = 0;
	if(dPos.x || dPos.y)
	{
		float dPosAngleRad = PI+atan2(dPos.y, dPos.x);
		float dPosAngle = RadToDeg(dPosAngleRad);
		if(PrintVector)
		{
			printf("%d %d %4.2f %4.2f\n", x,y, dPos.x ,dPos.y);
		}
		if(dPosAngle >= 22.0f && dPosAngle < 157.0f)
		{
			yOffset =  -1; // 0 is up
		}
		else if(dPosAngle >= 202.0f && dPosAngle < 330.0f)
		{
			yOffset = 1;
		}
		if((dPosAngle >= 292.0f && dPosAngle < 360.0f) ||
		   (dPosAngle >= 0.0f   && dPosAngle < 22.0f ))
		{
			xOffset =  1;
		}
		else if(dPosAngle >= 112.0f && dPosAngle < 257.0f)
		{
			xOffset =  -1;
		}
	}
#if 0
	if(xOffset!= 0 && yOffset != 0)
	{
		printf("Count :%f, v(%f,%f), Type: %d\n", Cell->ParticleCount, Cell->Velocity.x, Cell->Velocity.y, Cell->Type);
		printf("A: %f, %f \n", Acceleration.x, Acceleration.y);
		printf("xo = %d, yo = %d\n", xo,yo);
	}
#endif
	// Diagonal Check
	sim_cell *DNX = Cells + (SIZE*(y+yOffset))+x;
	sim_cell *DNY = Cells + (SIZE*y) + x + xOffset;
	// If there is any velocity
	if(xOffset!= 0 || yOffset != 0)
	{
		// TODO: Refactor this!!!
	        // if we are on the current cell
		if(xo == 0 && yo == 0)
		{
#if 1
				sim_cell *Cur = Cells + (y*SIZE) + x;
				sim_cell *NewCur = TempCells + ((y-1)*SIZE) + x;

				sim_cell *Neighbour = Cells + ((y+yo)*SIZE)+(x+xo);
				if(NewCur->Type == SimCellType_Space && Neighbour->Type == SimCellType_Space && (DNX->Type == SimCellType_Space || DNY->Type == SimCellType_Space))
				{
					float OutgoingParticleAmount = Cell->ParticleCount*Percentage;
				
					NewCur->ParticleCount -= OutgoingParticleAmount;
				}
#endif
		}
		// if we are on a neighbour cell
		else
		{
#if 1
		  //printf("Calc in %f\n", Cell->ParticleCount);		
			// Only add what is going into the current cell
			if(xOffset+x+xo==x && y+yo+yOffset==y)
			{
					sim_cell *Cur = Cells + (y*SIZE) + x;
					sim_cell *NewCur = TempCells + ((y-1)*SIZE) + x;

					sim_cell *Neighbour = Cells + ((y+yo)*SIZE) + (x+xo);
					if(NewCur->Type == SimCellType_Space && Cur->Type == SimCellType_Space && (DNX->Type == SimCellType_Space || DNY->Type == SimCellType_Space))
					{
					
						float IncommingParticleAmount = Neighbour->ParticleCount*Percentage;

						float TotalParticleCount = NewCur->ParticleCount + IncommingParticleAmount;
						//printf("Calc in %f\n", Percentage);
						// Avaraging the Velocity in based on the relative amount of particles
						if(TotalParticleCount != 0)
						{
							float InvTotalAmount = 1.0f/(TotalParticleCount);
							v2 OldPart = scale((InvTotalAmount*NewCur->ParticleCount),NewCur->Velocity);
							v2 NewPart = scale((InvTotalAmount*IncommingParticleAmount),NewVelocity);
							// Only one process should ever access this cell at a time. So this should be safe.
							NewCur->Velocity = add(OldPart,NewPart);
						}
						NewCur->ParticleCount += IncommingParticleAmount;
					}

			}
#endif
		}
}
}


//////////////////////////////////////////////////////////////////////////////////////////////////

// Loop for Testing
#if 0
 	for (int y = 1; y < RowsPerProcess+1 ; y++)
	{
		for (int x = 0; x < SIZE; x++)
		{
		        sim_cell *Cur = WorkArray + (y*SIZE) + x;
			sim_cell *TempCur = TempArray + ((y-1)*SIZE) + x;
			CopyCell(Cur, TempCur);
		        for(int xOffset = -1; xOffset <= 1 ; xOffset ++)
			{
				for(int yOffset = -1; yOffset <= 1 ; yOffset ++)
				{
					if(!(xOffset == 0 && yOffset == 0) &&
					   !(x+xOffset < 0 || x+xOffset >= SIZE) &&
					   !(y+yOffset < 0 || y+yOffset > RowsPerProcess+1) &&
					   !(y+yOffset+(ProcessRank*RowsPerProcess) < 1 || y+yOffset+(ProcessRank*RowsPerProcess) > SIZE))
					{
						sim_cell *Cell = (TempArray + ((y-1)*SIZE)+x);
						sim_cell *NeighborCell = (WorkArray + ((y+yOffset)*SIZE) +(x+xOffset));
						Cell->Velocity.x += NeighborCell->Velocity.x;
					}
				}
			}
		}
	}

#endif
